from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional

from pydantic import BaseModel
from pydantic import validator
from sqlmodel import Field  # type: ignore
from sqlmodel import SQLModel


__all__ = (
    "TaskCreate",
    "TaskUpdate",
    "TaskRead",
    "TaskImport",
    "TaskExport",
    "TaskCollectPip",
    "TaskCollectStatus",
)


class _TaskBase(SQLModel):
    """
    Task base class

    A Task is the elemental unit of a workflow, and must be a self-standing
    executable.

    Attributes
        name:
            A human readable name for the task
        command:
            The command(s) that executes the task
        source:
            Path or url to task source. This is the information is used to
            match tasks across fractal installations when a workflow is
            imported.
        input_type:
            The type of data the task expects as input
        output_type:
            The type of data the task expects as output
        default_args: Optional[Dict[str, Any]]
            Dictionary (saved as JSON) of the default parameters of the task
    """

    name: str
    command: str
    source: str
    input_type: str
    output_type: str
    default_args: Optional[Dict[str, Any]] = Field(default={})
    meta: Optional[Dict[str, Any]] = Field(default={})

    class Config:
        arbitrary_types_allowed = True


class TaskUpdate(_TaskBase):
    name: Optional[str]  # type:ignore
    input_type: Optional[str]  # type:ignore
    output_type: Optional[str]  # type:ignore
    command: Optional[str]  # type:ignore
    source: Optional[str]  # type:ignore
    default_args: Optional[Dict[str, Any]]  # type:ignore
    meta: Optional[Dict[str, Any]]  # type:ignore


class TaskImport(_TaskBase):
    pass


class TaskExport(_TaskBase):
    pass


class TaskRead(_TaskBase):
    id: int


class TaskCreate(_TaskBase):
    pass


class _TaskCollectBase(BaseModel):
    pass


class TaskCollectPip(_TaskCollectBase):
    """
    TaskCollectPip class

    Attributes:
        package: TBD
        version: TBD
        python_version: TBD
        package_extras: TBD
    """

    package: str
    version: Optional[str]
    python_version: Optional[str] = None
    package_extras: Optional[str]

    @validator("package")
    def package_path_absolute(cls, value):
        if "/" in value:
            package_path = Path(value)
            if not package_path.is_absolute():
                raise ValueError(
                    f"Package path must be absolute: {package_path}"
                )
            if not package_path.exists():
                raise ValueError(
                    f"Package file does not exist: {package_path}"
                )
        return value


class TaskCollectStatus(_TaskCollectBase):
    """
    TaskCollectStatus class

    Attributes:
        status: TBD
        package: TBD
        venv_path: TBD
        task_list: TBD
        log: TBD
        info: TBD
    """

    status: Literal["pending", "installing", "collecting", "fail", "OK"]
    package: str
    venv_path: Path
    task_list: Optional[List[TaskRead]] = Field(default=[])
    log: Optional[str]
    info: Optional[str]

    def sanitised_dict(self):
        """
        Return `self.dict()` after casting `self.venv_path` to a string
        """
        d = self.dict()
        d["venv_path"] = str(self.venv_path)
        return d



from .applyworkflow import *  # noqa: F403
from .manifest import *  # noqa: F403
from .project import *  # noqa: F403
from .state import *  # noqa: F403
from .task import *  # noqa: F403
from .user import *  # noqa: F403
from .workflow import *  # noqa: F403


__all__ = (
    project.__all__  # noqa: F405
    + task.__all__  # noqa: F405
    + workflow.__all__  # noqa: F405
    + applyworkflow.__all__  # noqa: F405
    + manifest.__all__  # noqa: F405
    + state.__all__  # noqa: F405
    + user.__all__  # noqa: F405
)



from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from pydantic import validator
from sqlmodel import SQLModel

from .task import TaskExport
from .task import TaskImport
from .task import TaskRead


__all__ = (
    "WorkflowCreate",
    "WorkflowRead",
    "WorkflowUpdate",
    "WorkflowImport",
    "WorkflowExport",
    "WorkflowTaskCreate",
    "WorkflowTaskImport",
    "WorkflowTaskExport",
    "WorkflowTaskRead",
    "WorkflowTaskUpdate",
)


class _WorkflowTaskBase(SQLModel):
    order: Optional[int]
    meta: Optional[Dict[str, Any]] = None
    args: Optional[Dict[str, Any]] = None


class WorkflowTaskCreate(_WorkflowTaskBase):
    task_id: int
    workflow_id: Optional[int]


class WorkflowTaskRead(_WorkflowTaskBase):
    id: int
    workflow_id: int
    task_id: int
    task: TaskRead


class WorkflowTaskImport(_WorkflowTaskBase):
    task: TaskImport


class WorkflowTaskExport(_WorkflowTaskBase):
    task: TaskExport


class WorkflowTaskUpdate(_WorkflowTaskBase):
    @validator("meta")
    def check_no_parallelisation_level(cls, m):
        if "parallelization_level" in m:
            raise ValueError(
                "Overriding task parallelization level currently not allowed"
            )
        return m


class _WorkflowBase(SQLModel):
    name: str


class WorkflowRead(_WorkflowBase):
    id: int
    project_id: int
    task_list: List[WorkflowTaskRead]


class WorkflowCreate(_WorkflowBase):
    project_id: int


class WorkflowUpdate(_WorkflowBase):
    name: Optional[str]  # type: ignore


class WorkflowImport(_WorkflowBase):
    task_list: List[WorkflowTaskImport]


class WorkflowExport(_WorkflowBase):
    task_list: List[WorkflowTaskExport]



from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import TypeVar

from pydantic import BaseModel
from pydantic import validator
from sqlmodel import Field


__all__ = ("TaskManifestV1", "ManifestV1")


class _TaskManifestBase(BaseModel):
    """
    Base class for TaskManifest

    Represents a task within a manfest

    Attributes:
        name:
            The task name
        executable:
            Path to the executable relative to the package root

            Note: by package root we mean "as it will be installed". If a
            package `Pkg` installs in the folder `pkg` the executable
            `pkg/executable.py`, this attribute must contain only
            `executable.py`.
        input_type:
            The input type accepted by the task
        output_type:
            The output type returned by the task
        default_args:
            An arbitrary, JSON-serializable dictionary containing the default
            parameters that will be passed to the task
        meta:
            Additional information about the package, such as hash of the
            executable, specific runtime requirements (e.g., need_gpu=True),
            etc.
    """

    name: str
    executable: Path
    input_type: str
    output_type: str
    default_args: Optional[Dict[str, Any]] = Field(default_factory=dict)
    meta: Optional[Dict[str, Any]] = Field(default_factory=dict)


TaskManifestType = TypeVar("TaskManifestType", bound=_TaskManifestBase)


class _ManifestBase(BaseModel):
    """
    Manifest base class

    Packages containing tasks are required to include a special file
    `__FRACTAL_MANIFEST__.json` in order to be discovered and used by Fractal.

    This model class and the model classes it depends on provide the base
    schema to read, write and validate manifests.

    Attributes:
        manifest_version:
            A version string that provides indication for compatibility between
            manifests as the schema evolves. This is for instance used by
            Fractal to determine which subclass of the present base class needs
            be used to read and validate the input.
        task_list : List[TaskManifestType]
            The list of tasks, represented as specified by subclasses of the
            _TaskManifestBase (a.k.a. TaskManifestType)
    """

    manifest_version: str
    task_list: List[TaskManifestType]  # type: ignore


class TaskManifestV1(_TaskManifestBase):
    pass


class ManifestV1(_ManifestBase):
    """
    Manifest schema version 1
    """

    task_list: List[TaskManifestV1]

    @validator("manifest_version")
    def manifest_version_1(cls, value):
        if value != "1":
            raise ValueError("Wrong manifest version")



from datetime import datetime
from typing import Any
from typing import Dict
from typing import Optional

from sqlmodel import SQLModel

__all__ = (
    "_StateBase",
    "StateRead",
)


class _StateBase(SQLModel):
    """
    Base class for `State`

    Attributes:
        id: Primary key
        data: Content of the state
        timestamp: Time stamp of the state
    """
    id: Optional[int]
    data: Dict[str, Any]
    timestamp: datetime

    class Config:
        arbitrary_types_allowed = True

    def sanitised_dict(self):
        """
        Return `self.dict()` after casting `self.timestamp` to a string
        """
        d = self.dict()
        d["timestamp"] = str(self.timestamp)
        return d


class StateRead(_StateBase):
    pass



from pathlib import Path

import jwt
from httpx import AsyncClient
from jwt.exceptions import ExpiredSignatureError

from .config import settings


class AuthenticationError(ValueError):
    pass


class AuthToken:
    def __init__(
        self,
        client: AsyncClient,
        username: str,
        password: str,
    ):
        self.client = client
        self.username = username
        self.password = password

        try:
            with open(f"{settings.FRACTAL_CACHE_PATH}/session", "r") as f:
                self.token = f.read()
        except FileNotFoundError:
            pass

    async def _get_fresh_token(self):
        data = dict(
            username=self.username,
            password=self.password,
        )
        res = await self.client.post(
            f"{settings.FRACTAL_SERVER}/auth/token/login", data=data
        )
        if res.status_code != 200:
            raise AuthenticationError(
                "Error: could not obtain token. Is the user registered?\n"
                f"{res.json()}\n"
            )
        raw_token = res.json()
        self.token = raw_token["access_token"]

        # Create cache folder, if needed
        cache_dir = Path(f"{settings.FRACTAL_CACHE_PATH}").expanduser()
        cache_dir.mkdir(parents=True, exist_ok=True)

        # Write token in cache_file
        cache_file = cache_dir / "session"
        with cache_file.open("w") as f:
            f.write(self.token)

    @property
    def expired(self):
        try:
            jwt.decode(
                jwt=self.token,
                requires=["exp"],
                options={
                    "verify_signature": False,
                    "verify_exp": True,
                },
            )
            return False
        except AttributeError:
            return True
        except ExpiredSignatureError:
            return True

    async def header(self):
        token = await self.__call__()
        return dict(Authorization=f"Bearer {token}")

    async def __call__(self):
        if self.expired:
            await self._get_fresh_token()
        return self.token


class AuthClient:
    def __init__(
        self,
        username: str,
        password: str,
    ):
        self.auth = None
        self.client = None
        self.username = username
        self.password = password

    async def __aenter__(self):
        self.client = AsyncClient()
        self.auth = AuthToken(
            client=self.client,
            username=self.username,
            password=self.password,
        )
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.client.aclose()

    async def get(self, *args, **kwargs):
        return await self.client.get(
            headers=await self.auth.header(), *args, **kwargs
        )

    async def post(self, *args, **kwargs):
        return await self.client.post(
            headers=await self.auth.header(), *args, **kwargs
        )

    async def patch(self, *args, **kwargs):
        return await self.client.patch(
            headers=await self.auth.header(), *args, **kwargs
        )

    async def delete(self, *args, **kwargs):
        return await self.client.delete(
            headers=await self.auth.header(), *args, **kwargs
        )



from httpx import AsyncClient

from ..authclient import AuthClient
from ..config import __VERSION__
from ..config import settings
from ..interface import BaseInterface
from ..interface import PrintInterface
from ._dataset import dataset_add_resource
from ._dataset import dataset_delete
from ._dataset import dataset_delete_resource
from ._dataset import dataset_edit
from ._dataset import dataset_show
from ._job import job_download_logs
from ._job import job_list
from ._job import job_show
from ._project import project_add_dataset
from ._project import project_create
from ._project import project_delete
from ._project import project_list
from ._project import project_show
from ._task import task_collect_pip
from ._task import task_collection_check
from ._task import task_delete
from ._task import task_edit
from ._task import task_list
from ._task import task_new
from ._user import user_delete
from ._user import user_edit
from ._user import user_list
from ._user import user_register
from ._user import user_show
from ._user import user_whoami
from ._workflow import workflow_add_task
from ._workflow import workflow_apply
from ._workflow import workflow_delete
from ._workflow import workflow_edit
from ._workflow import workflow_edit_task
from ._workflow import workflow_export
from ._workflow import workflow_import
from ._workflow import workflow_list
from ._workflow import workflow_new
from ._workflow import workflow_remove_task
from ._workflow import workflow_show


class NoCommandError(ValueError):
    pass


async def project(
    client: AuthClient, subcmd: str, batch: bool = False, **kwargs
) -> BaseInterface:
    if subcmd == "new":
        iface = await project_create(client, batch=batch, **kwargs)
    elif subcmd == "show":
        iface = await project_show(client, **kwargs)
    elif subcmd == "list":
        iface = await project_list(client, **kwargs)
    elif subcmd == "add-dataset":
        iface = await project_add_dataset(
            client,
            batch=batch,
            project_id=kwargs.pop("project_id"),
            dataset_name=kwargs.pop("dataset_name"),
            metadata_filename=kwargs.pop("metadata"),
            **kwargs,
        )
    elif subcmd == "delete":
        iface = await project_delete(client, **kwargs)
    else:
        raise NoCommandError(f"Command project {subcmd} not found")

    return iface


async def dataset(
    client: AuthClient, subcmd: str, batch: bool = False, **kwargs
) -> BaseInterface:
    if subcmd == "show":
        iface = await dataset_show(client, **kwargs)
    elif subcmd == "add-resource":
        iface = await dataset_add_resource(
            client,
            batch=batch,
            project_id=kwargs.pop("project_id"),
            dataset_id=kwargs.pop("dataset_id"),
            path=kwargs.pop("path"),
            glob_pattern=kwargs.pop("glob_pattern"),
            **kwargs,
        )
    elif subcmd == "rm-resource":
        iface = await dataset_delete_resource(
            client,
            batch=batch,
            project_id=kwargs.pop("project_id"),
            dataset_id=kwargs.pop("dataset_id"),
            resource_id=kwargs.pop("resource_id"),
            **kwargs,
        )
    elif subcmd == "edit":
        project_id = int(kwargs.pop("project_id"))
        dataset_id = int(kwargs.pop("dataset_id"))

        dataset_update_dict = kwargs
        iface = await dataset_edit(
            client,
            project_id=project_id,
            dataset_id=dataset_id,
            dataset_update_dict=dataset_update_dict,
        )
    elif subcmd == "delete":
        iface = await dataset_delete(client, **kwargs)
    else:
        raise NoCommandError(f"Command dataset {subcmd} not found")
    return iface


async def task(
    client: AuthClient, subcmd: str, batch: bool = False, **kwargs
) -> BaseInterface:
    if subcmd == "list":
        iface = await task_list(client, **kwargs)
    elif subcmd == "collect":
        iface = await task_collect_pip(client, batch=batch, **kwargs)
    elif subcmd == "check-collection":
        iface = await task_collection_check(client, **kwargs)
    elif subcmd == "new":
        iface = await task_new(client, batch=batch, **kwargs)
    elif subcmd == "edit":
        iface = await task_edit(client, **kwargs)
    elif subcmd == "delete":
        iface = await task_delete(client, **kwargs)
    else:
        raise NoCommandError(f"Command task {subcmd} not found")
    return iface


async def workflow(
    client: AuthClient, subcmd: str, batch: bool = False, **kwargs
) -> BaseInterface:
    if subcmd == "show":
        iface = await workflow_show(client, **kwargs)
    elif subcmd == "new":
        iface = await workflow_new(client, batch=batch, **kwargs)
    elif subcmd == "list":
        iface = await workflow_list(client, batch=batch, **kwargs)
    elif subcmd == "edit":
        iface = await workflow_edit(client, **kwargs)
    elif subcmd == "delete":
        iface = await workflow_delete(client, **kwargs)
    elif subcmd == "add-task":
        iface = await workflow_add_task(client, batch=batch, **kwargs)
    elif subcmd == "edit-task":
        iface = await workflow_edit_task(client, **kwargs)
    elif subcmd == "rm-task":
        iface = await workflow_remove_task(client, **kwargs)
    elif subcmd == "apply":
        iface = await workflow_apply(client, **kwargs)
    elif subcmd == "import":
        iface = await workflow_import(client, batch=batch, **kwargs)
    elif subcmd == "export":
        iface = await workflow_export(client, **kwargs)
    else:
        raise NoCommandError(f"Command workflow {subcmd} not found")
    return iface


async def job(
    client: AuthClient, subcmd: str, batch: bool = False, **kwargs
) -> BaseInterface:
    if subcmd == "list":
        iface = await job_list(client, batch=batch, **kwargs)
    elif subcmd == "show":
        iface = await job_show(client, batch=batch, **kwargs)
    elif subcmd == "download-logs":
        iface = await job_download_logs(client, **kwargs)
    else:
        raise NoCommandError(f"Command job {subcmd} not found")
    return iface


async def version(client: AsyncClient, **kwargs) -> PrintInterface:
    res = await client.get(f"{settings.FRACTAL_SERVER}/api/alive/")
    data = res.json()

    return PrintInterface(
        retcode=0,
        data=(
            f"Fractal client\n\tversion: {__VERSION__}\n"
            "Fractal server:\n"
            f"\turl: {settings.FRACTAL_SERVER}"
            f"\tdeployment type: {data['deployment_type']}"
            f"\tversion: {data['version']}"
        ),
    )


async def user(
    client: AuthClient, subcmd: str, batch: bool = False, **kwargs
) -> BaseInterface:
    if subcmd == "register":
        iface = await user_register(client, batch=batch, **kwargs)
    elif subcmd == "list":
        iface = await user_list(client, **kwargs)
    elif subcmd == "show":
        iface = await user_show(client, **kwargs)
    elif subcmd == "edit":
        iface = await user_edit(client, **kwargs)
    elif subcmd == "delete":
        iface = await user_delete(client, **kwargs)
    elif subcmd == "whoami":
        iface = await user_whoami(client, batch=batch, **kwargs)
    else:
        raise NoCommandError(f"Command user {subcmd} not found")

    return iface



import json
import logging
from pathlib import Path
from typing import Optional

from ..authclient import AuthClient
from ..common.schemas import ApplyWorkflowCreate
from ..common.schemas import ApplyWorkflowRead
from ..common.schemas import WorkflowCreate
from ..common.schemas import WorkflowExport
from ..common.schemas import WorkflowImport
from ..common.schemas import WorkflowRead
from ..common.schemas import WorkflowTaskCreate
from ..common.schemas import WorkflowTaskRead
from ..common.schemas import WorkflowTaskUpdate
from ..common.schemas import WorkflowUpdate
from ..config import settings
from ..interface import BaseInterface
from ..interface import PrintInterface
from ..interface import RichJsonInterface
from ..response import check_response
from .utils import get_cached_task_by_name


async def workflow_new(
    client: AuthClient,
    name: str,
    project_id: int,
    batch: bool = False,
    **kwargs,
) -> BaseInterface:
    workflow = WorkflowCreate(
        name=name,
        project_id=project_id,
    )
    logging.info(workflow)
    res = await client.post(
        f"{settings.BASE_URL}/workflow/",
        json=workflow.dict(),
    )
    workflow = check_response(
        res, expected_status_code=201, coerce=WorkflowRead
    )
    if batch:
        return PrintInterface(retcode=0, data=workflow.id)
    else:
        return RichJsonInterface(retcode=0, data=workflow.dict())


async def workflow_list(
    client: AuthClient,
    project_id: int,
    batch: bool = False,
    **kwargs,
) -> RichJsonInterface:

    res = await client.get(
        f"{settings.BASE_URL}/project/{project_id}/workflows/"
    )
    workflow_list = check_response(res, expected_status_code=200)
    return RichJsonInterface(retcode=0, data=workflow_list)


async def workflow_delete(
    client: AuthClient,
    *,
    workflow_id: int,
    **kwargs,
) -> BaseInterface:
    res = await client.delete(f"{settings.BASE_URL}/workflow/{workflow_id}")
    check_response(res, expected_status_code=204)
    return PrintInterface(retcode=0, data="")


async def workflow_show(
    client: AuthClient,
    *,
    workflow_id: int,
    **kwargs,
) -> RichJsonInterface:
    res = await client.get(f"{settings.BASE_URL}/workflow/{workflow_id}")
    workflow = check_response(
        res, expected_status_code=200, coerce=WorkflowRead
    )
    return RichJsonInterface(retcode=0, data=workflow.dict())


async def workflow_add_task(
    client: AuthClient,
    *,
    workflow_id: int,
    batch: bool = False,
    task_id_or_name: str,
    order: int = None,
    args_file: Optional[str] = None,
    meta_file: Optional[str] = None,
    **kwargs,
) -> RichJsonInterface:

    try:
        task_id = int(task_id_or_name)
    except ValueError:
        task_id = await get_cached_task_by_name(task_id_or_name, client)

    workflow_task = WorkflowTaskCreate(task_id=task_id, order=order)
    if args_file:
        with Path(args_file).open("r") as f:
            args = json.load(f)
            workflow_task.args = args
    if meta_file:
        with Path(meta_file).open("r") as f:
            meta = json.load(f)
            workflow_task.meta = meta

    res = await client.post(
        f"{settings.BASE_URL}/workflow/{workflow_id}/add-task/",
        json=workflow_task.dict(),
    )
    workflow_task = check_response(
        res, expected_status_code=201, coerce=WorkflowTaskRead
    )

    if batch:
        return PrintInterface(retcode=0, data=str(workflow_task.id))
    else:
        return RichJsonInterface(retcode=0, data=workflow_task.dict())


async def workflow_edit_task(
    client: AuthClient,
    *,
    workflow_id: int,
    workflow_task_id: int,
    args_file: Optional[str] = None,
    meta_file: Optional[str] = None,
    **kwargs,
) -> RichJsonInterface:

    # Check that at least one of args_file or meta_file was given (note: it
    # would be reasonable to check it in the parser, but we are not aware of a
    # method within argparse).
    if not (args_file or meta_file):
        raise ValueError(
            "At least one of {args_file,meta_file} arguments is required"
        )

    # Combine args/meta payload
    payload = {}
    if args_file:
        with Path(args_file).open("r") as f:
            args = json.load(f)
            payload["args"] = args
    if meta_file:
        with Path(meta_file).open("r") as f:
            meta = json.load(f)
            payload["meta"] = meta

    payload_update = WorkflowTaskUpdate(**payload)
    res = await client.patch(
        f"{settings.BASE_URL}/"
        f"workflow/{workflow_id}/edit-task/{workflow_task_id}",
        json=payload_update.dict(exclude_unset=True),
    )
    workflow_task = check_response(
        res, expected_status_code=200, coerce=WorkflowTaskRead
    )

    return RichJsonInterface(retcode=0, data=workflow_task.dict())


async def workflow_remove_task(
    client: AuthClient,
    *,
    workflow_id: int,
    workflow_task_id: int,
    **kwargs,
) -> BaseInterface:

    res = await client.delete(
        f"{settings.BASE_URL}/"
        f"workflow/{workflow_id}/rm-task/{workflow_task_id}"
    )
    check_response(res, expected_status_code=204)
    return PrintInterface(retcode=0, data="")


async def workflow_edit(
    client: AuthClient,
    *,
    workflow_id: int,
    **workflow_update_dict,
) -> BaseInterface:
    workflow_update = WorkflowUpdate(**workflow_update_dict)
    payload = workflow_update.dict(exclude_unset=True)
    if not payload:
        return PrintInterface(retcode=1, data="Nothing to update")

    res = await client.patch(
        f"{settings.BASE_URL}/workflow/{workflow_id}", json=payload
    )
    new_workflow = check_response(
        res, expected_status_code=200, coerce=WorkflowRead
    )
    return RichJsonInterface(retcode=0, data=new_workflow.dict())


async def workflow_apply(
    client: AuthClient,
    *,
    workflow_id: int,
    input_dataset_id: int,
    output_dataset_id: int,
    # output_dataset_id: Optional[int] = None,
    overwrite_input: bool = False,
    project_id: Optional[int] = None,
    worker_init: Optional[str] = None,
    **kwargs,
) -> BaseInterface:
    apply_wf_create = ApplyWorkflowCreate(
        workflow_id=workflow_id,
        input_dataset_id=input_dataset_id,
        output_dataset_id=output_dataset_id,
        overwrite_input=overwrite_input,
        project_id=project_id,
        worker_init=worker_init,
    )

    res = await client.post(
        f"{settings.BASE_URL}/project/apply/", json=apply_wf_create.dict()
    )
    apply_wf_read = check_response(
        res, expected_status_code=202, coerce=ApplyWorkflowRead
    )
    return RichJsonInterface(retcode=0, data=apply_wf_read.sanitised_dict())


async def workflow_import(
    client: AuthClient,
    *,
    project_id: int,
    json_file: str,
    batch: bool = False,
    **kwargs,
) -> BaseInterface:
    with Path(json_file).open("r") as f:
        workflow = json.load(f)
    workflow = WorkflowImport(**workflow)
    res = await client.post(
        f"{settings.BASE_URL}/project/{project_id}/import-workflow/",
        json=workflow.dict(),
    )
    wf_read = check_response(
        res, expected_status_code=201, coerce=WorkflowRead
    )
    if batch:
        datastr = f"{wf_read.id}"
        for wftask in wf_read.task_list:
            datastr += f" {wftask.id}"
        return PrintInterface(retcode=0, data=datastr)
    else:
        return RichJsonInterface(retcode=0, data=wf_read.dict())


async def workflow_export(
    client: AuthClient,
    *,
    workflow_id: int,
    json_file: str,
    **kwargs,
) -> BaseInterface:
    res = await client.get(
        f"{settings.BASE_URL}/workflow/{workflow_id}/export/"
    )
    workflow = check_response(
        res, expected_status_code=200, coerce=WorkflowExport
    )
    with Path(json_file).open("w") as f:
        json.dump(workflow.dict(), f, indent=2)
    return PrintInterface(
        retcode=0, data=f"Workflow {workflow_id} exported at {json_file}"
    )



import logging
import os
from pathlib import Path
from zipfile import ZipFile

from rich.table import Table

from ..authclient import AuthClient
from ..common.schemas import ApplyWorkflowRead
from ..config import settings
from ..interface import BaseInterface
from ..interface import PrintInterface
from ..interface import RichConsoleInterface
from ..interface import RichJsonInterface
from ..response import check_response


async def job_show(
    client: AuthClient,
    job_id: int,
    batch: bool = False,
    do_not_separate_logs: bool = False,
    **kwargs,
) -> BaseInterface:
    """
    Query the status of a workflow-execution job
    """

    res = await client.get(f"{settings.BASE_URL}/job/{job_id}")
    job = check_response(
        res, expected_status_code=200, coerce=ApplyWorkflowRead
    )
    if batch:
        return PrintInterface(retcode=0, data=job.status)
    else:
        data = job.sanitised_dict()
        if do_not_separate_logs or (job.log is None):
            return RichJsonInterface(retcode=0, data=data)
        else:
            log = data.pop("log")
            extra_lines = ["\nThis is the job log:\n", log]
            return RichJsonInterface(
                retcode=0, data=data, extra_lines=extra_lines
            )


async def job_list(
    client: AuthClient,
    project_id: int,
    batch: bool = False,
    **kwargs,
) -> BaseInterface:

    res = await client.get(f"{settings.BASE_URL}/project/{project_id}/jobs/")
    jobs = check_response(res, expected_status_code=200)

    # Coerce to a list of ApplyWorkflowRead objects
    jobs = [ApplyWorkflowRead(**job) for job in jobs]

    if batch:
        job_ids = " ".join(str(job.id) for job in jobs)
        return PrintInterface(retcode=0, data=job_ids)
    else:
        table = Table(title=f"Job list for project {project_id}")
        kwargs = dict(style="white", justify="center")
        table.add_column("id", **kwargs)
        table.add_column("start_timestamp", **kwargs)
        table.add_column("status", **kwargs)
        table.add_column("workflow_id", **kwargs)
        table.add_column("working_dir", **kwargs)

        for j in jobs:
            timestamp = j.dict()["start_timestamp"]
            table.add_row(
                str(j.id),
                timestamp.strftime("%Y-%m-%d %H:%M:%S"),
                j.status,
                str(j.workflow_id),
                j.working_dir,
            )
        return RichConsoleInterface(retcode=0, data=table)


async def job_download_logs(
    client: AuthClient,
    job_id: int,
    output_folder: str,
    **kwargs,
) -> BaseInterface:

    # Check that output_folder does not already exist
    if Path(output_folder).exists():
        return PrintInterface(
            retcode=1, data=f"ERROR: {output_folder=} already exists"
        )

    # Send request to server
    res = await client.get(f"{settings.BASE_URL}/job/download/{job_id}")

    # NOTE: We cannot use our default check_response here, because res._content
    # is binary. Therefore we check the status code by hand
    if res.status_code != 200:
        logging.error(f"Server returned {res.status_code}")
        logging.error(
            f"Original request: {res._request.method} {res._request.url}"
        )
        logging.error(
            f"Original payload: {res._request._content.decode('utf-8')}"
        )
        logging.error("Terminating.\n")
        exit(1)

    # Check the content-type entry in the response headers
    content_type = res.headers["content-type"]
    expected_content_type = "application/x-zip-compressed"
    if content_type != expected_content_type:
        logging.error(
            f"Unexpected {content_type=} in headers of server "
            f"response, instead of {expected_content_type=}"
        )
        logging.error(
            f"Original request: {res._request.method} {res._request.url}"
        )
        logging.error(
            f"Original payload: {res._request._content.decode('utf-8')}"
        )
        logging.error("Terminating.\n")
        exit(1)

    # Write response into a temporary zipped file
    zipped_archive_path = output_folder + "_tmp.zip"
    with open(zipped_archive_path, "wb") as f:
        f.write(res.content)

    # Unzip the log archive
    unzipped_archived_path = output_folder
    os.mkdir(unzipped_archived_path)
    with ZipFile(zipped_archive_path, mode="r") as zipfile:
        zipfile.extractall(path=unzipped_archived_path)

    # Remove zipped temporary file
    os.unlink(zipped_archive_path)

    return PrintInterface(
        retcode=0, data=f"Logs downloaded to {output_folder=}"
    )



from typing import Optional
from typing import Union

from ..authclient import AuthClient
from ..common.schemas.user import UserCreate
from ..common.schemas.user import UserRead
from ..common.schemas.user import UserUpdate
from ..config import settings
from ..interface import PrintInterface
from ..interface import RichJsonInterface
from ..response import check_response


async def user_register(
    client: AuthClient,
    new_email: str,
    new_password: Optional[str] = None,
    slurm_user: Optional[str] = None,
    superuser: bool = False,
    batch: bool = False,
    **kwargs,
) -> Union[RichJsonInterface, PrintInterface]:

    new_user = UserCreate(
        email=new_email, password=new_password, slurm_user=slurm_user
    )

    from getpass import getpass

    if new_password is None:
        new_password = getpass()
        confirm_new_password = getpass("Confirm password: ")
        if new_password == confirm_new_password:
            new_user.password = new_password
        else:
            return PrintInterface(retcode=1, data="Passwords do not match.")

    res = await client.post(
        f"{settings.FRACTAL_SERVER}/auth/register",
        json=new_user.dict(exclude_unset=True, exclude_none=True),
    )
    data = check_response(res, expected_status_code=201, coerce=UserRead)

    if superuser:
        user_id = data.id
        res = await client.patch(
            f"{settings.FRACTAL_SERVER}/auth/users/{user_id}",
            json={"is_superuser": True},
        )
        data = check_response(res, expected_status_code=200, coerce=UserRead)

    if batch:
        return PrintInterface(retcode=0, data=data.id)
    else:
        return RichJsonInterface(retcode=0, data=data.dict())


async def user_list(client: AuthClient, **kwargs) -> RichJsonInterface:
    res = await client.get(f"{settings.FRACTAL_SERVER}/auth/userlist")
    users = check_response(res, expected_status_code=200)
    users = [UserRead(**user).dict() for user in users]
    return RichJsonInterface(
        retcode=0,
        data=users,
    )


async def user_show(
    client: AuthClient, user_id: str, **kwargs
) -> RichJsonInterface:
    res = await client.get(f"{settings.FRACTAL_SERVER}/auth/users/{user_id}")
    user = check_response(res, expected_status_code=200, coerce=UserRead)
    return RichJsonInterface(
        retcode=0,
        data=user.dict(),
    )


async def user_edit(
    client: AuthClient,
    user_id: str,
    new_email: Optional[str] = None,
    new_password: Optional[str] = None,
    new_slurm_user: Optional[str] = None,
    make_superuser: bool = False,
    remove_superuser: bool = False,
    **kwargs,
) -> Union[RichJsonInterface, PrintInterface]:

    user_update = UserUpdate(
        email=new_email,
        password=new_password,
        slurm_user=new_slurm_user,
    )

    if make_superuser:
        user_update.is_superuser = True
    if remove_superuser:
        user_update.is_superuser = False

    payload = user_update.dict(exclude_unset=True, exclude_none=True)
    if not payload:
        return PrintInterface(retcode=1, data="Nothing to update")

    res = await client.patch(
        f"{settings.FRACTAL_SERVER}/auth/users/{user_id}", json=payload
    )
    new_user = check_response(res, expected_status_code=200, coerce=UserRead)

    return RichJsonInterface(
        retcode=0,
        data=new_user.dict(),
    )


async def user_delete(
    client: AuthClient, user_id: str, **kwargs
) -> PrintInterface:

    res = await client.delete(
        f"{settings.FRACTAL_SERVER}/auth/users/{user_id}"
    )
    check_response(res, expected_status_code=204)
    return PrintInterface(retcode=0, data="")


async def user_whoami(
    client: AuthClient, batch: bool, **kwargs
) -> Union[RichJsonInterface, PrintInterface]:
    res = await client.get(f"{settings.FRACTAL_SERVER}/auth/whoami")
    user = check_response(res, expected_status_code=200, coerce=UserRead)

    if batch:
        return PrintInterface(retcode=0, data=user.id)
    else:
        return RichJsonInterface(retcode=0, data=user.dict())



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Jacopo Nespolo <jacopo.nespolo@exact-lab.it>
Tommaso Comparin <tommaso.comparin@exact-lab.it>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
import logging
from os import getenv
from typing import Optional

from dotenv import load_dotenv
from pydantic import BaseSettings

from . import __VERSION__


def fail_getenv(key):
    value = getenv(key, None)
    if value is None:
        raise ValueError(f"Must provide environment variable {key}")
    return value


load_dotenv(".fractal.env")


class Settings(BaseSettings):
    PROJECT_NAME: str = "Fractal client"
    PROJECT_VERSION: str = __VERSION__

    FRACTAL_LOGGING_LEVEL: int = logging.INFO

    FRACTAL_USER: Optional[str] = getenv("FRACTAL_USER")
    FRACTAL_PASSWORD: Optional[str] = getenv("FRACTAL_PASSWORD")

    FRACTAL_SERVER: str = getenv("FRACTAL_SERVER", "http://localhost:8000")

    BASE_URL: str = f"{FRACTAL_SERVER}/api/v1"
    FRACTAL_CACHE_PATH: str = "~/.cache/fractal"


settings = Settings()



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Jacopo Nespolo <jacopo.nespolo@exact-lab.it>
Marco Franzon <marco.franzon@exact-lab.it>
Tommaso Comparin <tommaso.comparin@exact-lab.it>
Yuri Chiucconi  <yuri.chiucconi@exact-lab.it>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
import argparse as ap


parser_main = ap.ArgumentParser(
    description="Command-line interface for Fractal Client"
)

parser_main.add_argument(
    "-u",
    "--user",
    help="User email address for login (overrides configuration file)",
)
parser_main.add_argument(
    "-p", "--password", help="User password (overrides configuration file)"
)
parser_main.add_argument(
    "--verbose",
    "-v",
    action="store_true",
    default=False,
    help="Change minimal logging level from default to DEBUG",
)
parser_main.add_argument(
    "--batch",
    default=False,
    action="store_true",
    help=(
        "Return output suitable for scripting, e.g., "
        "only the ID of items created instead of the full object."
    ),
)

subparsers_main = parser_main.add_subparsers(title="Commands:", dest="cmd")


# PROJECT GROUP
project_parser = subparsers_main.add_parser(
    "project", description="Project commands"
)
project_subparsers = project_parser.add_subparsers(
    title="Valid subcommands:", dest="subcmd", required=True
)

# project new
project_new_parser = project_subparsers.add_parser(
    "new", description="Create new project"
)
project_new_parser.add_argument("name", help="Name of new project", type=str)
project_new_parser.add_argument(
    "path",
    help=(
        "Project directory of new project. "
        "New datasets will be written here."
    ),
)
project_new_parser.add_argument(
    "-d",
    "--dataset",
    help=(
        "Name of new dataset to create. "
        "By default, the dataset `default` is created."
    ),
)

# project list
project_list_parser = project_subparsers.add_parser(
    "list", description="List projects"
)

# project show
project_show_parser = project_subparsers.add_parser(
    "show", description="Show details of single project"
)
project_show_parser.add_argument(
    "project_id", type=int, help="ID of project to show"
)

# project delete
project_delete_parser = project_subparsers.add_parser(
    "delete", description="Delete project"
)
project_delete_parser.add_argument(
    "project_id", type=int, help="ID of project to delete"
)

# project add-dataset
project_add_dataset_parser = project_subparsers.add_parser(
    "add-dataset", description="Add dataset to project"
)
project_add_dataset_parser.add_argument(
    "project_id", type=int, help="ID of project to add the new dataset to"
)
project_add_dataset_parser.add_argument(
    "dataset_name", help="Name of new dataset"
)
project_add_dataset_parser.add_argument(
    "--metadata",
    help="Path to file containing dataset metadata in JSON format.",
)


# DATASET GROUP
dataset_parser = subparsers_main.add_parser(
    "dataset", description="Dataset commands"
)
dataset_subparsers = dataset_parser.add_subparsers(
    title="Valid subcommands:", dest="subcmd", required=True
)

# dataset add-resource
dataset_add_resource_parser = dataset_subparsers.add_parser(
    "add-resource", description="Add resource to existing dataset"
)
dataset_add_resource_parser.add_argument(
    "project_id", type=int, help="Project ID"
)
dataset_add_resource_parser.add_argument(
    "dataset_id", type=int, help="Dataset ID"
)
dataset_add_resource_parser.add_argument("path", help="Path to resource")
dataset_add_resource_parser.add_argument(
    "-g", "--glob-pattern", help="Glob pattern"
)

# dataset rm-resource
dataset_rm_resource_parser = dataset_subparsers.add_parser(
    "rm-resource", description="Remove resource to existing dataset"
)
dataset_rm_resource_parser.add_argument(
    "project_id", type=int, help="Project ID"
)
dataset_rm_resource_parser.add_argument(
    "dataset_id", type=int, help="Dataset ID"
)
dataset_rm_resource_parser.add_argument(
    "resource_id", type=int, help="Resource ID"
)

# dataset edit
dataset_edit_parser = dataset_subparsers.add_parser(
    "edit", description="Edit dataset", argument_default=ap.SUPPRESS
)
dataset_edit_parser.add_argument("project_id", type=int, help="Project ID")
dataset_edit_parser.add_argument("dataset_id", type=int, help="Dataset ID")
dataset_edit_parser.add_argument("--name", help="New name of dataset")
dataset_edit_parser.add_argument("--path", help="New path of dataset")
dataset_edit_parser.add_argument(
    "--metadata",
    help=(
        "Path to file containing dataset metadata in JSON format. "
        "(Set to `none` to clear)"
    ),
)
dataset_edit_parser.add_argument(
    "--read-only",
    dest="read_only",
    help="Set read-only flag of dataset",
    action="store_true",
)
dataset_edit_parser.add_argument(
    "--read-write",
    dest="read_only",
    help="Set read-only flag of dataset (0 for False, 1 for True)",
    action="store_false",
)
dataset_edit_parser.add_argument("-t", "--type", help="Dataset type")

# dataset show
dataset_show_parser = dataset_subparsers.add_parser(
    "show", description="Show dataset", argument_default=ap.SUPPRESS
)
dataset_show_parser.add_argument("project_id", type=int, help="Project ID")
dataset_show_parser.add_argument("dataset_id", type=int, help="Dataset ID")

# dataset delete
dataset_delete_parser = dataset_subparsers.add_parser(
    "delete", description="Delete dataset", argument_default=ap.SUPPRESS
)
dataset_delete_parser.add_argument("project_id", type=int, help="Project ID")
dataset_delete_parser.add_argument("dataset_id", type=int, help="Dataset ID")


# TASK GROUP
task_parser = subparsers_main.add_parser("task", description="Task commands")
task_subparsers = task_parser.add_subparsers(
    title="Valid subcommands:", dest="subcmd", required=True
)

# task list
task_list_parser = task_subparsers.add_parser("list", description="List tasks")

# task collect
task_collect_parser = task_subparsers.add_parser(
    "collect",
    description="Install and collect all tasks from a pip-installable package",
)
task_collect_parser.add_argument(
    "package",
    help="Package name or path to local package",
)
task_collect_parser.add_argument(
    "--python-version",
    help="Select the python version to use for this package",
)
task_collect_parser.add_argument(
    "--package-version",
    help="Select the package version",
)
task_collect_parser.add_argument(
    "--package-extras",
    help=(
        "Comma separated list of extra components for the package to be "
        "installed, e.g., `collect fractal-tasks-core "
        "--package-extras=torch,tensorflow` will trigger the installation of "
        "`fractal-tasks-core[torch,tensorflow]`"
    ),
)
task_collect_parser.add_argument(
    "--private",
    default=False,
    action="store_true",
    help="Intall tasks as private to the user (as opposed to global)",
)

# task check-collection
task_check_collection_parser = task_subparsers.add_parser(
    "check-collection",
    description="Check status of background task collection processes",
)
task_check_collection_parser.add_argument(
    "state_id",
    help="State ID of the collection (see output of task collect)",
)
task_check_collection_parser.add_argument(
    "--include-logs",
    default=False,
    action="store_true",
    help="Also include task-collection logs",
)

# task new
task_new_parser = task_subparsers.add_parser(
    "new", description="Create new task", argument_default=ap.SUPPRESS
)
task_new_parser.add_argument(
    "name", help="A human readable name for the task", type=str
)
task_new_parser.add_argument(
    "command", help="The command that executes the task", type=str
)
task_new_parser.add_argument("source", help="TBD", type=str)
task_new_parser.add_argument(
    "--input-type",
    help="The type of data the task expects as input",
    type=str,
    default="Any",
)
task_new_parser.add_argument(
    "--output-type",
    help="The type of data the task expects as output",
    type=str,
    default="Any",
)
task_new_parser.add_argument(
    "--default-args-file",
    help="Path to JSON file with default task arguments",
    type=str,
)
task_new_parser.add_argument(
    "--meta-file",
    help="Path to JSON file with additional parameters useful for execution",
    type=str,
)

# task edit
task_edit_parser = task_subparsers.add_parser(
    "edit", description="Edit task", argument_default=ap.SUPPRESS
)
task_edit_parser.add_argument(
    "task_id_or_name", help="ID or name of task to edit", type=str
)
task_edit_parser.add_argument("--name", help="New task name")
task_edit_parser.add_argument(
    "--input-type",
    help="New input type",
)
task_edit_parser.add_argument(
    "--output-type",
    help="New resource type",
)
task_edit_parser.add_argument(
    "--default-args", help="Filename containing JSON encoded default arguments"
)

# task delete
task_delete_parser = task_subparsers.add_parser(
    "delete", description="Delete task", argument_default=ap.SUPPRESS
)
task_delete_parser.add_argument(
    "task_id_or_name", help="ID or name of task to delete", type=str
)


# WORKFLOW GROUP

workflow_parser = subparsers_main.add_parser(
    "workflow", description="Workflow commands"
)
workflow_subparsers = workflow_parser.add_subparsers(
    title="Valid subcommand", dest="subcmd", required=True
)

# workflow list
workflow_list_parser = workflow_subparsers.add_parser(
    "list", description="List workflows for given project"
)
workflow_list_parser.add_argument(
    "project_id",
    help="Project ID",
)

# workflow new
workflow_new_parser = workflow_subparsers.add_parser(
    "new", description="Create new workflow"
)
workflow_new_parser.add_argument(
    "name",
    help="Workflow name (must be unique, and not only made of numbers only)",
)
workflow_new_parser.add_argument(
    "project_id",
    help="Project ID",
)

# workflow show
workflow_new_parser = workflow_subparsers.add_parser(
    "show", description="Show workflow"
)
workflow_new_parser.add_argument(
    "workflow_id",
    help="Workflow ID",
)

# workflow edit
workflow_edit_parser = workflow_subparsers.add_parser(
    "edit", description="Edit workflow", argument_default=ap.SUPPRESS
)
workflow_edit_parser.add_argument(
    "workflow_id",
    type=int,
    help="Workflow ID",
)
workflow_edit_parser.add_argument("--name", type=str, help="New workflow name")

# workflow delete
workflow_delete_parser = workflow_subparsers.add_parser(
    "delete", description="Delete workflow"
)
workflow_delete_parser.add_argument(
    "workflow_id",
    help="Workflow ID",
)


# workflow add task
workflow_add_task_parser = workflow_subparsers.add_parser(
    "add-task", description="Add new task to specific workflow"
)
workflow_add_task_parser.add_argument(
    "workflow_id",
    help="Workflow ID",
)
workflow_add_task_parser.add_argument(
    "task_id_or_name", help="ID or name of the new task", type=str
)
workflow_add_task_parser.add_argument(
    "--order", help="Order of this task within the workflow's task list"
)
workflow_add_task_parser.add_argument(
    "--args-file",
    help=(
        "Path to json serialised file containing the arguments "
        "ovverrides of the task"
    ),
)
workflow_add_task_parser.add_argument(
    "--meta-file",
    help=(
        "Path to json serialised file containing the meta"
        "ovverrides of the task"
    ),
)

# workflow edit task
workflow_edit_task_parser = workflow_subparsers.add_parser(
    "edit-task", description="Edit task within specific workflow"
)
workflow_edit_task_parser.add_argument(
    "workflow_id",
    help="Workflow ID",
)
workflow_edit_task_parser.add_argument(
    "workflow_task_id",
    help="Workflow task ID, the ID of a task inside the list of tasks",
)
workflow_edit_task_parser.add_argument(
    "--args-file",
    help=(
        "Path to json serialised file containing the arguments "
        "ovverrides of the task"
    ),
)
workflow_edit_task_parser.add_argument(
    "--meta-file",
    help=(
        "Path to json serialised file containing the meta "
        "ovverrides of the task"
    ),
)

# workflow remove task
workflow_remove_task_parser = workflow_subparsers.add_parser(
    "rm-task", description="Remove task from a specific workflow"
)
workflow_remove_task_parser.add_argument(
    "workflow_id",
    help="Workflow ID",
)
workflow_remove_task_parser.add_argument(
    "workflow_task_id",
    help="Workflow task ID (the ID of a task inside the list of tasks)",
)

# workflow apply
workflow_apply_parser = workflow_subparsers.add_parser(
    "apply",
    description="Apply workflow to dataset",
    argument_default=ap.SUPPRESS,
)
workflow_apply_parser.add_argument("workflow_id")
workflow_apply_parser.add_argument("input_dataset_id")
workflow_apply_parser.add_argument(
    "-o", "--output_dataset_id", help="Output dataset ID"
)
workflow_apply_parser.add_argument(
    "--overwrite-input",
    default=False,
    action="store_true",
    help="Allow overwriting the content of the input dataset",
)
workflow_apply_parser.add_argument(
    "-p",
    "--project-id",
    help="ID of project the workflow and dataset belong to",
)
workflow_apply_parser.add_argument(
    "-w",
    "--worker-init",
    help="Command to be run before starting a worker",
)

# workflow import
workflow_import_parser = workflow_subparsers.add_parser(
    "import",
    description="Import workflow to project from file",
)
workflow_import_parser.add_argument(
    "--project-id",
    help="ID of the project where the workflow will be imported",
    required=True,
)
workflow_import_parser.add_argument(
    "--json-file",
    help="Path to a JSON file with the workflow to be imported",
)

# workflow export
workflow_export_parser = workflow_subparsers.add_parser(
    "export",
    description="Export workflow to file",
)
workflow_export_parser.add_argument(
    "workflow_id",
    help="Workflow ID",
)
workflow_export_parser.add_argument(
    "--json-file",
    help="Path to the JSON file where the workflow will be exported",
    required=True,
)

# JOB GROUP

job_parser = subparsers_main.add_parser("job", description="Job commands")
job_subparsers = job_parser.add_subparsers(
    title="Valid subcommand", dest="subcmd", required=True
)

# job list
job_list_parser = job_subparsers.add_parser(
    "list", description="List jobs for given project"
)
job_list_parser.add_argument(
    "project_id",
    help="Project ID",
)

# job show
job_show_parser = job_subparsers.add_parser(
    "show",
    description="Query status of workflow-execution job",
    argument_default=ap.SUPPRESS,
)
job_show_parser.add_argument(
    "job_id",
    help="ID of the job",
)
job_show_parser.add_argument(
    "--do-not-separate-logs",
    dest="do_not_separate_logs",
    help="Show the job logs in the main output, instead of a separate field",
    action="store_true",
    required=False,
)

# job download-logs
job_download_logs_parser = job_subparsers.add_parser(
    "download-logs",
    description="Download full folder of workflow-execution job",
)
job_download_logs_parser.add_argument(
    "job_id",
    help="ID of the job",
)
job_download_logs_parser.add_argument(
    "--output",
    dest="output_folder",
    help="Path of the output folder",
    required=True,
)


# VERSION GROUP
version_parser = subparsers_main.add_parser(
    "version", description="Print version and exit"
)


# USER GROUP

user_parser = subparsers_main.add_parser(
    "user",
    description="User commands",
)
user_subparsers = user_parser.add_subparsers(
    title="Valid subcommand", dest="subcmd", required=True
)

# user whoami
user_whoami_parser = user_subparsers.add_parser(
    "whoami",
    description="Get info on current user (fails if user is not registered)",
)

# user register
user_register_parser = user_subparsers.add_parser(
    "register", description="Register a new user with the Fractal server"
)
user_register_parser.add_argument(
    "new_email", help="Email to be used as username"
)
user_register_parser.add_argument(
    "new_password", help="Password for the new user"
)
user_register_parser.add_argument(
    "--slurm-user",
    help="Username to login into SLURM cluster",
    required=False,
)
user_register_parser.add_argument(
    "--superuser",
    help="Give superuser privileges to the new user",
    action="store_true",
    required=False,
)

# user list
user_list_parser = user_subparsers.add_parser(
    "list",
    description="List all users",
)

# user show
user_show_parser = user_subparsers.add_parser(
    "show", description="Show details of single user"
)
user_show_parser.add_argument("user_id", help="ID of the user")

# user edit
user_edit_parser = user_subparsers.add_parser(
    "edit", description="Edit details of single user"
)
user_edit_parser.add_argument("user_id", help="ID of the user")
user_edit_parser.add_argument(
    "--new-email", help="New email address", type=str, required=False
)
user_edit_parser.add_argument(
    "--new-password", help="New password", type=str, required=False
)
user_edit_parser.add_argument(
    "--new-slurm-user", help="New SLURM username", type=str, required=False
)

user_edit_parser_superuser = user_edit_parser.add_mutually_exclusive_group()
user_edit_parser_superuser.add_argument(
    "--make-superuser",
    help="Give superuser privileges to user",
    action="store_true",
    required=False,
)
user_edit_parser_superuser.add_argument(
    "--remove-superuser",
    help="Remove superuser privileges from user",
    action="store_true",
    required=False,
)

# user delete
user_delete_parser = user_subparsers.add_parser(
    "delete", description="Delete a single user"
)
user_delete_parser.add_argument("user_id", help="ID of the user")



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

from rich import print_json
from rich.console import Console


class BaseInterface:
    def __init__(self, retcode: int, data=None):
        self.retcode = retcode
        self.data = data

    def show(self, *args, **kwargs):
        raise NotImplementedError("Implement in subclasses")

    def __repr__(self):
        return f"<Interface (retcode={self.retcode})>"


class PrintInterface(BaseInterface):
    """
    Basic interface that simply prints a string to screen
    """

    def __init__(self, retcode: int, data: Any):
        super().__init__(retcode, data)

    def show(self, *args, **kwargs):
        print(str(self.data))


class RichJsonInterface(BaseInterface):
    """
    Output json using rich.print_json
    """

    def __init__(
        self,
        retcode: int,
        data: Union[Dict[str, Any], List[Dict[str, Any]]],
        extra_lines: Optional[Sequence[str]] = None,
    ):
        super().__init__(retcode, data)
        if extra_lines is None:
            self.extra_lines = None
        else:
            self.extra_lines = "\n".join(extra_lines)

    def show(self, *args, **kwargs):
        print_json(data=self.data)
        if self.extra_lines:
            print(self.extra_lines)


class RichConsoleInterface(BaseInterface):
    def __init__(self, retcode: int, data: Any):
        super().__init__(retcode, data)

    def show(self, *args, **kwargs):
        console = Console()
        console.print(self.data)



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Jacopo Nespolo <jacopo.nespolo@exact-lab.it>
Marco Franzon <marco.franzon@exact-lab.it>
Tommaso Comparin <tommaso.comparin@exact-lab.it>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
import asyncio
import logging
from sys import argv
from typing import List
from typing import Tuple
from typing import Union

from httpx import AsyncClient
from httpx import ConnectError

from . import cmd
from .authclient import AuthClient
from .authclient import AuthenticationError
from .config import settings
from .interface import PrintInterface
from .parser import parser_main


class MissingCredentialsError(RuntimeError):
    pass


def _check_credentials(
    *, username: Union[str, None], password: Union[str, None]
) -> Tuple[str, str]:
    """
    Check that username and password are defined

    Arguments:
        username: Username
        password: Password

    Raises:
        MissingCredentialsError: If either `username` of `password` is `None`.
    """
    if not username:
        message = (
            "FRACTAL_USER variable not defined."
            "\nPossible options: \n"
            + "    1. Set --user argument;\n"
            + "    2. Define FRACTAL_USER in a .fractal.env file;\n"
            + "    3. Define FRACTAL_USER as an environment variable."
        )
        raise MissingCredentialsError(message)
    if not password:
        message = (
            "FRACTAL_PASSWORD variable not defined."
            "\nPossible options: \n"
            + "    1. Set --password argument;\n"
            + "    2. Define FRACTAL_PASSWORD in a .fractal.env file;\n"
            + "    3. Define FRACTAL_PASSWORD as an environment variable."
        )
        raise MissingCredentialsError(message)
    return (username, password)


async def handle(cli_args: List[str] = argv):
    args = parser_main.parse_args(cli_args[1:])

    # Set logging level
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=settings.FRACTAL_LOGGING_LEVEL)

    logging.debug(f"Arguments: {args}")

    if args.cmd:
        handler = getattr(cmd, args.cmd)
    else:
        # no command provided. Print help and exit 1
        parser_main.print_help()
        exit(1)

    try:
        kwargs = vars(args)
        if args.cmd == "version":
            async with AsyncClient() as client:
                interface = await handler(client, **kwargs)
        else:
            # Extract (and remove) username/password for AuthClient from kwargs
            username = kwargs.pop("user") or settings.FRACTAL_USER
            password = kwargs.pop("password") or settings.FRACTAL_PASSWORD
            username, password = _check_credentials(
                username=username, password=password
            )
            async with AuthClient(
                username=username, password=password
            ) as client:
                interface = await handler(client, **kwargs)
    except AuthenticationError as e:
        return PrintInterface(retcode=1, data=e.args[0])
    except ConnectError as e:
        return PrintInterface(retcode=1, data=e.args[0])

    return interface


async def main():
    interface = await handle()
    interface.show()
    exit(interface.retcode)


if __name__ == "__main__":
    asyncio.run(main())



from datetime import datetime
from typing import List
from typing import Optional

from sqlmodel import SQLModel


__all__ = (
    "ApplyWorkflowBase",
    "ApplyWorkflowCreate",
    "ApplyWorkflowRead",
)


class ApplyWorkflowBase(SQLModel):
    """
    Base class for ApplyWorkflow

    Attributes:
        input_dataset_id: TBD
        output_dataset_id: TBD
        workflow_id: TBD
        overwrite_input: TBD
        worker_init: TBD
        working_dir: TBD
        working_dir_user: TBD
    """

    project_id: int
    input_dataset_id: int
    output_dataset_id: Optional[int]
    workflow_id: int
    overwrite_input: bool = False
    worker_init: Optional[str]
    working_dir: Optional[str]
    working_dir_user: Optional[str]


class ApplyWorkflowCreate(ApplyWorkflowBase):
    pass


class ApplyWorkflowRead(ApplyWorkflowBase):
    id: int
    start_timestamp: datetime
    status: str
    log: Optional[str]
    history: Optional[List[str]]

    def sanitised_dict(self):
        d = self.dict()
        d["start_timestamp"] = str(self.start_timestamp)
        return d



import uuid
from typing import Optional

from fastapi_users import schemas

__all__ = (
    "UserRead",
    "UserUpdate",
    "UserCreate",
)


class UserRead(schemas.BaseUser[uuid.UUID]):
    slurm_user: Optional[str]


class UserUpdate(schemas.BaseUserUpdate):
    slurm_user: Optional[str]


class UserCreate(schemas.BaseUserCreate):
    slurm_user: Optional[str]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fractal', 'fractal.cmd', 'fractal.common', 'fractal.common.schemas']

package_data = \
{'': ['*']}

install_requires = \
['PyJWT>=2.4.0,<3.0.0',
 'fastapi-users[oauth]>=10.1,<11.0',
 'httpx>=0.23.0,<0.24.0',
 'pydantic>=1.10.1,<2.0.0',
 'python-dotenv>=0.20.0,<0.21.0',
 'rich>=12.5.1,<13.0.0',
 'sqlmodel>=0.0.8,<0.0.9']

entry_points = \
{'console_scripts': ['fractal = fractal.__main__:run',
                     'fractal_new = fractal.new:run',
                     'fractal_old = fractal.old:run']}

setup_kwargs = {
    'name': 'fractal-client',
    'version': '1.0.5',
    'description': 'Client component of the Fractal analytics platform',
    'long_description': '# Fractal Client\n\n[![PyPI version](https://img.shields.io/pypi/v/fractal-client?color=gree)](https://pypi.org/project/fractal-client/)\n[![CI Status](https://github.com/fractal-analytics-platform/fractal/actions/workflows/ci.yml/badge.svg)](https://github.com/fractal-analytics-platform/fractal/actions/workflows/ci.yml)\n[![Documentation Status](https://github.com/fractal-analytics-platform/fractal/actions/workflows/documentation.yaml/badge.svg)](https://fractal-analytics-platform.github.io/fractal)\n[![License](https://img.shields.io/badge/License-BSD_3--Clause-blue.svg)](https://opensource.org/licenses/BSD-3-Clause)\n\nFractal is a framework to process high-content imaging data at scale and prepare it for interactive visualization.\n\n![Fractal_Overview](https://fractal-analytics-platform.github.io/assets/fractal_overview.jpg)\n\nFractal provides distributed workflows that convert TBs of image data into OME-Zarr files. The platform then processes the 3D image data by applying tasks like illumination correction, maximum intensity projection, 3D segmentation using [cellpose](https://cellpose.readthedocs.io/en/latest/) and measurements using [napari workflows](https://github.com/haesleinhuepf/napari-workflows). The pyramidal OME-Zarr files enable interactive visualization in the napari viewer.\n\nThis is the repository that contains the **Fractal client**. Find more information about Fractal in general and the other repositories at the [Fractal home page](https://fractal-analytics-platform.github.io).\n\n## Documentation\n\nSee https://fractal-analytics-platform.github.io/fractal.\n\n# Contributors and license\n\nUnless otherwise stated in each individual module, all Fractal components are released according to a BSD 3-Clause License, and Copyright is with Friedrich Miescher Institute for Biomedical Research and University of Zurich.\n\nFractal was conceived in the Liberali Lab at the Friedrich Miescher Institute for Biomedical Research and in the Pelkmans Lab at the University of Zurich (both in Switzerland). The project lead is with [@gusqgm](https://github.com/gusqgm) & [@jluethi](https://github.com/jluethi). The core development is done under contract by [@mfranzon](https://github.com/mfranzon), [@tcompa](https://github.com/tcompa) & [@jacopo-exact](https://github.com/jacopo-exact) from [eXact lab S.r.l.](exact-lab.it).\n',
    'author': 'Tommaso Comparin',
    'author_email': 'tommaso.comparin@exact-lab.it',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/fractal-analytics-platform/fractal',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import json
import logging
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List

from ..authclient import AuthClient
from ..config import settings
from ..response import check_response


async def get_cached_task_by_name(name: str, client: AuthClient) -> int:

    # Set paths
    cache_dir = Path(f"{settings.FRACTAL_CACHE_PATH}").expanduser()
    cache_file = cache_dir / "tasks"

    # If cache is missing, create it
    cache_up_to_date = False
    if not cache_file.exists():
        await refresh_task_cache(client)
        cache_up_to_date = True

    # Read cache
    with cache_file.open("r") as f:
        task_cache = json.load(f)

    # Look for name in cache
    # Case 1: name exists in cache
    if name in task_cache.keys():
        return task_cache[name]
    # Case 2: name is missing, and cache was just updated
    elif cache_up_to_date:
        raise KeyError(f'Task "{name}" not in {cache_file}\n')
    # Case 3: name is missing but cache may be out of date
    else:
        await refresh_task_cache(client)
        with cache_file.open("r") as f:
            task_cache = json.load(f)
        try:
            return task_cache[name]
        except KeyError as e:
            raise KeyError(f'Task "{name}" not in {cache_file}\n', str(e))


async def refresh_task_cache(
    client: AuthClient, **kwargs
) -> List[Dict[str, Any]]:

    # Get task_list
    res = await client.get(f"{settings.BASE_URL}/task/")
    task_list = check_response(res, expected_status_code=200)

    # Check that there are no name clashes
    names = [task["name"] for task in task_list]
    if len(set(names)) < len(names):
        msg = (
            "Cannot write task-list cache if task names are not unique "
            "(version-based disambiguation will be added in the future).\n"
            f"Current task list includes: {names}"
        )
        logging.warning(msg)
        return task_list

    # Refresh cache of (name,id) pairs
    task_cache = {}
    for task in task_list:
        task_cache[task["name"]] = task["id"]

    # Set paths and create cache folder (if needed)
    cache_dir = Path(f"{settings.FRACTAL_CACHE_PATH}").expanduser()
    cache_file = cache_dir / "tasks"
    cache_dir.mkdir(parents=True, exist_ok=True)

    # Write task cache
    with cache_file.open("w") as f:
        json.dump(task_cache, f, indent=4)

    return task_list



import logging
from json.decoder import JSONDecodeError
from sys import exit


def check_response(res, expected_status_code=200, coerce=False):
    """
    Check the validity of the http response from fractal server

    If the status code of the response is not one of the expected values, print
    the error to stderr and terminate with exit status 1.

    On success, optionally coerce to a pydantic model
    """

    # Also allow a list of expected status codes
    if isinstance(expected_status_code, list):
        expected_status_codes = expected_status_code
    else:
        expected_status_codes = [expected_status_code]

    logging.debug(res.status_code)
    try:
        data = res.json()
    except JSONDecodeError:
        data = {}
    if res.status_code not in expected_status_codes:
        logging.error(f"Server returned {res.status_code}")
        logging.error(
            f"Original request: {res._request.method} {res._request.url}"
        )
        logging.error(
            f"Original payload: {res._request._content.decode('utf-8')}"
        )
        logging.error(f"Server error message: {data}\n")
        logging.error("Terminating.\n")
        exit(1)

    if coerce:
        return coerce(**data)
    else:
        return data
